@@ -7,7 +7,6 @@
 #include "snappy/snappy.h"
 
 #include "miniparquet.h"
-#include "bpacking.h"
 
 #include <protocol/TCompactProtocol.h>
 #include <transport/TBufferTransports.h>
@@ -329,188 +328,37 @@ private:
 		return true;
 	}
 
-	// TODO this is slow because it always starts from scratch, reimplement with state
+	// somewhat optimized implementation that avoids non-alignment
+
+	struct bitpack_state {
+		uint64_t bitpack_off = 0;
+		int8_t bitpack_pos = 0;
+	};
+
+	static const uint32_t BITPACK_MASKS[];
+	static const uint8_t BITPACK_DLEN;
+
 	template<typename T>
-	static T bitunpack_rev(const uint8_t *source, uint64_t *source_offset,
-			uint8_t encoding_length) {
-
-		T target = 0;
-		for (auto j = 0; j < encoding_length; j++, (*source_offset)++) {
-			target |= (1 & (source[(*source_offset) / 8] >> *source_offset % 8))
-					<< j;
+	static T bitunpack_rev2(const uint8_t *source, uint8_t encoding_length,
+			bitpack_state *state) {
+		assert(encoding_length < 32);
+		T val = (source[state->bitpack_off] >> state->bitpack_pos)
+				& BITPACK_MASKS[encoding_length];
+		state->bitpack_pos += encoding_length;
+		while (state->bitpack_pos > BITPACK_DLEN) {
+			val |= (source[++state->bitpack_off]
+					<< (BITPACK_DLEN - (state->bitpack_pos - encoding_length)))
+					& BITPACK_MASKS[encoding_length];
+			state->bitpack_pos -= BITPACK_DLEN;
 		}
-		return target;
+		return val;
 	}
 
-	// from Lemire
-	static int unpack32(uint32_t *in, uint32_t *out, int batch_size,
-			int num_bits) {
-		batch_size = batch_size / 32 * 32;
-		int num_loops = batch_size / 32;
-
-		int base = 0;
-
-		switch (num_bits) {
-		case 0:
-			for (int i = 0; i < num_loops; ++i)
-				in = nullunpacker32(base, in, out + i * 32);
-			break;
-		case 1:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack1_32(base, in, out + i * 32);
-			break;
-		case 2:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack2_32(base, in, out + i * 32);
-			break;
-		case 3:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack3_32(base, in, out + i * 32);
-			break;
-		case 4:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack4_32(base, in, out + i * 32);
-			break;
-		case 5:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack5_32(base, in, out + i * 32);
-			break;
-		case 6:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack6_32(base, in, out + i * 32);
-			break;
-		case 7:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack7_32(base, in, out + i * 32);
-			break;
-		case 8:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack8_32(base, in, out + i * 32);
-			break;
-		case 9:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack9_32(base, in, out + i * 32);
-			break;
-		case 10:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack10_32(base, in, out + i * 32);
-			break;
-		case 11:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack11_32(base, in, out + i * 32);
-			break;
-		case 12:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack12_32(base, in, out + i * 32);
-			break;
-		case 13:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack13_32(base, in, out + i * 32);
-			break;
-		case 14:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack14_32(base, in, out + i * 32);
-			break;
-		case 15:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack15_32(base, in, out + i * 32);
-			break;
-		case 16:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack16_32(base, in, out + i * 32);
-			break;
-		case 17:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack17_32(base, in, out + i * 32);
-			break;
-		case 18:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack18_32(base, in, out + i * 32);
-			break;
-		case 19:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack19_32(base, in, out + i * 32);
-			break;
-		case 20:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack20_32(base, in, out + i * 32);
-			break;
-		case 21:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack21_32(base, in, out + i * 32);
-			break;
-		case 22:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack22_32(base, in, out + i * 32);
-			break;
-		case 23:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack23_32(base, in, out + i * 32);
-			break;
-		case 24:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack24_32(base, in, out + i * 32);
-			break;
-		case 25:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack25_32(base, in, out + i * 32);
-			break;
-		case 26:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack26_32(base, in, out + i * 32);
-			break;
-		case 27:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack27_32(base, in, out + i * 32);
-			break;
-		case 28:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack28_32(base, in, out + i * 32);
-			break;
-		case 29:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack29_32(base, in, out + i * 32);
-			break;
-		case 30:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack30_32(base, in, out + i * 32);
-			break;
-		case 31:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack31_32(base, in, out + i * 32);
-			break;
-		case 32:
-			for (int i = 0; i < num_loops; ++i)
-				in = unpack32_32(base, in, out + i * 32);
-			break;
-		default:
-			throw runtime_error("Unsupported bit packing width");
-		}
-
-		return batch_size;
-	}
-
-	// XXX for now we only need 1 bit to one byte for levels and n bits to 32 bit for offsets
-	//
-	// TODO this needs to check whether there is enough buffer left
 	template<typename T>
 	uint32_t BitUnpack(T *dest, uint32_t count) {
-
-		if (sizeof(T) == 4) {
-			// the fast unpacker needs to read 32 values at a time
-			auto bitpack_read_size = ((count + 31) / 32) * 32;
-			unpack_buf.resize(sizeof(T) * bitpack_read_size, false);
-
-			unpack32((uint32_t*) buffer, (uint32_t*) unpack_buf.ptr,
-					bitpack_read_size, bit_width_);
-			memcpy(dest, unpack_buf.ptr, count * sizeof(T));
-
-		} else {
-			uint64_t buffer_offset = 0;
-			for (uint32_t i = 0; i < count; i++) {
-				dest[i] = bitunpack_rev<T>(buffer, &buffer_offset, bit_width_);
-			}
-
+		bitpack_state state;
+		for (uint32_t i = 0; i < count; i++) {
+			dest[i] = bitunpack_rev2<T>(buffer, bit_width_, &state);
 		}
 		buffer += bit_width_ * count / 8;
 		return count;
@@ -518,6 +366,13 @@ private:
 
 };
 
+const uint32_t RleBpDecoder::BITPACK_MASKS[] = { 0, 1, 3, 7, 15, 31, 63, 127,
+		255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 131071, 262143,
+		524287, 1048575, 2097151, 4194303, 8388607, 16777215, 33554431,
+		67108863, 134217727, 268435455, 536870911, 1073741823, 2147483647 };
+
+const uint8_t RleBpDecoder::BITPACK_DLEN = 8;
+
 class ColumnScan {
 public:
 	PageHeader page_header;
@@ -540,11 +395,15 @@ public:
 		auto dict_size = page_header.dictionary_page_header.num_values;
 		dict = new Dictionary<T>(dict_size);
 		for (int32_t dict_index = 0; dict_index < dict_size; dict_index++) {
-			((Dictionary<T>*) dict)->dict[dict_index] = *(T*) page_buf_ptr;
+			T val;
+			memcpy(&val, page_buf_ptr, sizeof(val));
 			page_buf_ptr += sizeof(T);
+
+			((Dictionary<T>*) dict)->dict[dict_index] = val;
 		}
 	}
 
+
 	void scan_dict_page(ResultColumn &result_col) {
 		if (page_header.__isset.data_page_header
 				|| !page_header.__isset.dictionary_page_header) {
@@ -601,8 +460,9 @@ public:
 			dict = new Dictionary<char*>(dict_size);
 
 			for (int32_t dict_index = 0; dict_index < dict_size; dict_index++) {
-				uint32_t str_len = *((uint32_t*) page_buf_ptr);
-				page_buf_ptr += sizeof(uint32_t);
+				uint32_t str_len;
+				memcpy(&str_len, page_buf_ptr, sizeof(str_len));
+				page_buf_ptr += sizeof(str_len);
 
 				if (page_buf_ptr + str_len > page_buf_end_ptr) {
 					throw runtime_error(
@@ -642,8 +502,9 @@ public:
 		switch (page_header.data_page_header.definition_level_encoding) {
 		case Encoding::RLE: {
 			// read length of define payload, always
-			auto def_length = *((uint32_t*) page_buf_ptr);
-			page_buf_ptr += sizeof(uint32_t);
+			uint32_t def_length;
+			memcpy(&def_length, page_buf_ptr, sizeof(def_length));
+			page_buf_ptr += sizeof(def_length);
 
 			RleBpDecoder dec((const uint8_t*) page_buf_ptr, def_length, 1);
 			dec.GetBatch<uint8_t>(defined_ptr, num_values);
@@ -685,14 +546,14 @@ public:
 			}
 
 			auto row_idx = page_start_row + val_offset;
-
-			result_arr[row_idx] = *((T*) page_buf_ptr);
+			T val;
+			memcpy(&val, page_buf_ptr, sizeof(val));
 			page_buf_ptr += sizeof(T);
+			result_arr[row_idx] = val;
 		}
 	}
 
 	void scan_data_page_plain(ResultColumn &result_col) {
-
 		// TODO compute null count while getting the def levels already?
 		uint32_t null_count = 0;
 		for (uint32_t i = 0; i < page_header.data_page_header.num_values; i++) {
@@ -702,10 +563,24 @@ public:
 		}
 
 		switch (result_col.col->type) {
-		case Type::BOOLEAN:
-			fill_values_plain<bool>(result_col);
-			break;
+		case Type::BOOLEAN: {
+			// uargh, but unfortunately neccessary because sometimes bool values are > 1
+			bool *result_arr = (bool*) result_col.data.ptr;
+			for (int32_t val_offset = 0;
+					val_offset < page_header.data_page_header.num_values;
+					val_offset++) {
 
+				if (!defined_ptr[val_offset]) {
+					continue;
+				}
+
+				auto row_idx = page_start_row + val_offset;
+				result_arr[row_idx] = ((bool*) page_buf_ptr != 0);
+				page_buf_ptr += sizeof(bool);
+			}
+
+		}
+			break;
 		case Type::INT32:
 			fill_values_plain<int32_t>(result_col);
 			break;
@@ -747,8 +622,8 @@ public:
 				auto row_idx = page_start_row + val_offset;
 
 				if (result_col.col->type == Type::BYTE_ARRAY) {
-					str_len = *((uint32_t*) page_buf_ptr);
-					page_buf_ptr += sizeof(uint32_t);
+					memcpy(&str_len, page_buf_ptr, sizeof(str_len));
+					page_buf_ptr += sizeof(str_len);
 				}
 
 				if (page_buf_ptr + str_len > page_buf_end_ptr) {
@@ -829,6 +704,7 @@ public:
 		}
 
 		switch (result_col.col->type) {
+		// TODO no bools here? I guess makes no sense to use dict...
 
 		case Type::INT32:
 			fill_values_dict<int32_t>(result_col, offsets.get());
@@ -876,6 +752,40 @@ public:
 		}
 	}
 
+
+	// ugly but well
+	void cleanup(ResultColumn &result_col) {
+		switch (result_col.col->type) {
+		case Type::BOOLEAN:
+			delete (Dictionary<bool>*) dict;
+			break;
+		case Type::INT32:
+			delete (Dictionary<int32_t>*) dict;
+			break;
+		case Type::INT64:
+			delete (Dictionary<int64_t>*) dict;
+			break;
+		case Type::INT96:
+			delete (Dictionary<Int96>*) dict;
+			break;
+		case Type::FLOAT:
+			delete (Dictionary<float>*) dict;
+			break;
+		case Type::DOUBLE:
+			delete (Dictionary<double>*) dict;
+			break;
+		case Type::BYTE_ARRAY:
+		case Type::FIXED_LEN_BYTE_ARRAY:
+			delete (Dictionary<char*>*) dict;
+			break;
+		default:
+			throw runtime_error(
+					"Unsupported type for dictionary: "
+							+ type_to_string(result_col.col->type));
+		}
+
+	}
+
 };
 
 void ParquetFile::scan_column(ScanState &state, ResultColumn &result_col) {
@@ -908,7 +818,7 @@ void ParquetFile::scan_column(ScanState &state, ResultColumn &result_col) {
 	// read entire chunk into RAM
 	pfile.seekg(chunk_start);
 	ByteBuffer chunk_buf;
-	chunk_buf.resize(chunk_len + 32 * sizeof(uint32_t)); // extra space at the back for efficient and safe bit packing decoding
+	chunk_buf.resize(chunk_len);
 
 	pfile.read(chunk_buf.ptr, chunk_len);
 	if (!pfile) {
@@ -957,7 +867,7 @@ void ParquetFile::scan_column(ScanState &state, ResultColumn &result_col) {
 			size_t decompressed_size;
 			snappy::GetUncompressedLength(chunk_buf.ptr,
 					cs.page_header.compressed_page_size, &decompressed_size);
-			decompressed_buf.resize(decompressed_size + 32 * sizeof(uint32_t)); // see above
+			decompressed_buf.resize(decompressed_size + 1);
 
 			auto res = snappy::RawUncompress(chunk_buf.ptr,
 					cs.page_header.compressed_page_size, decompressed_buf.ptr);
@@ -996,6 +906,7 @@ void ParquetFile::scan_column(ScanState &state, ResultColumn &result_col) {
 		chunk_buf.ptr = payload_end_ptr;
 		bytes_to_read -= cs.page_header.compressed_page_size;
 	}
+	cs.cleanup(result_col);
 }
 
 void ParquetFile::initialize_column(ResultColumn &col, uint64_t num_rows) {
@@ -1034,7 +945,7 @@ void ParquetFile::initialize_column(ResultColumn &col, uint64_t num_rows) {
 		if (!s_ele->__isset.type_length) {
 			throw runtime_error("need a type length for fixed byte array");
 		}
-		col.data.resize(num_rows * s_ele->type_length, false);
+		col.data.resize(num_rows * sizeof(char*), false);
 		break;
 	}
 
